unit CFXThreads;
//
// !!!!!!!!!!!!!!!!!!  Warning  !!!!!!!!!!!!!!!!!!!!!
// Cannot pass functions to either of the procedures in this unit that make calls to Delphi VCL or Windows GDI.
// Any multi threaded function that would need to take advantage of Synchronized calls (where the sub thread is
// suspended and the main thread calls that function) cannot be used.  A synchronized call would defeat the pur-
// pose of using these functions since we definetely want these functions to be called in a separate thread because
// it is a slow function.
//

interface

uses Classes;

type
  TProcPointer = procedure();
  TMethPointer = procedure() of object;
  TProcError =  procedure(const Arg: string);
  TMethError =  procedure(const Arg: string) of object;

  TThreadProc = class(TThread)
    private
      threadProc, notifyProc: TProcPointer;
      threadMeth, notifyMeth: TMethPointer;
      errorProc: TProcError;
      errorMeth: TMethError;
    protected
      procedure Execute; override;
    public
      constructor create(const thread, notify: TProcPointer; const err: TProcError); overload;
      constructor create(const thread, notify: TMethPointer; const err: TMethError); overload;
  end;

procedure runThreadedProc(const threadProc: TProcPointer; const notifyProc: TProcPointer = nil; const errProc: TProcError = nil);
procedure runThreadedMeth(const threadMeth: TMethPointer; const notifyMeth: TMethPointer = nil; const errMeth: TMethError = nil);

implementation

uses SysUtils;

procedure runThreadedProc(const threadProc: TProcPointer; const notifyProc: TProcPointer = nil; const errProc: TProcError = nil) stdcall;
begin

  //
  // Caller's function is run in create.  The object frees itself
  //
  TThreadProc.create(threadProc, notifyProc, errProc);

end;

procedure runThreadedMeth(const threadMeth: TMethPointer; const notifyMeth: TMethPointer = nil; const errMeth: TMethError = nil) stdcall;
begin

  //
  // Caller's function is run in create.  The object frees itself
  //
  TThreadProc.create(threadMeth, notifyMeth, errMeth);

end;

{ TThreadProc }

constructor TThreadProc.create(const thread, notify: TProcPointer; const err: TProcError);
begin
  inherited Create(true);

  self.threadProc := thread;
  self.notifyProc := notify;
  self.errorProc := err;
  self.threadMeth := nil;

  FreeOnTerminate := true;
  Resume;
end;

constructor TThreadProc.create(const thread, notify: TMethPointer; const err: TMethError);
begin
  inherited Create(true);

  self.threadMeth := thread;
  self.notifyMeth := notify;
  self.errorMeth := err;
  self.threadProc := nil;  

  FreeOnTerminate := true;
  Resume;
end;

procedure TThreadProc.Execute;
//
// Run the procedure passed in by caller
//
begin
  if assigned(threadMeth) then begin

    try
      threadMeth()
    except
      on err:Exception do
        if assigned(errorMeth) then
          errorMeth(err.Message);
    end;
    if assigned(notifyMeth) then
      notifyMeth();

  end else begin

    try
      threadProc()
    except
      on err:Exception do
        if assigned(errorProc) then
          errorProc(err.Message);
    end;
    if assigned(notifyProc) then
      notifyProc();

  end;
end;

end.
