unit PreciseTimer;

{
  TPreciseTimer Component
  Version: 1.0

  Date: 01/02/2004 (mm/dd/yyyy)

  Code by Tom Wetjens aka Wetgos
  E-mail: wetgos@wetgos.nl

  Powerful timing component to be used in games
  or other programs that require accurate timing.

  This component provides timing accurate to
  0.0001 seconds. The timer can trigger events
  at a stable speed, so that your game runs just as
  fast on the users machine as it did on your
  own machine.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TRealTimeEvent = procedure (Sender: TObject; Fps: Real) of object;

  TPreciseTimer = class(TComponent)
  private
    FTimeFreq: Int64;
    FTimeCurrent: Int64;
    FProcessLast, FRenderLast: Int64;
    FTimeOverhead: Int64;
    FDoProcess: Boolean;
    FDoRender: Boolean;
    FEnabled: Boolean;
    FDesiredRender: Double;
    FRealRender: Double;
    FDesiredProcess: Double;
    FRealProcess: Double;
    FOnProcess: TNotifyEvent;
    FOnRender: TNotifyEvent;
    FDoRealTime: Boolean;
    FOnRealTime: TRealTimeEvent;
    procedure SetDesiredProcess(const Value: Double);
    procedure SetDesiredRender(const Value: Double);
    procedure SetDoProcess(const Value: Boolean);
    procedure SetDoRender(const Value: Boolean);
    procedure SetEnabled(const Value: Boolean);
    procedure SetOnProcess(const Value: TNotifyEvent);
    procedure SetOnRender(const Value: TNotifyEvent);
    procedure SetDoRealTime(const Value: Boolean);
    procedure SetOnRealTime(const Value: TRealTimeEvent);
  protected
    procedure ApplicationIdle(Sender: TObject; var Done: Boolean);
  public
    property RealRender: Double read FRealRender;
    property RealProcess: Double read FRealProcess;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property DesiredRender: Double read FDesiredRender write SetDesiredRender;
    property DesiredProcess: Double read FDesiredProcess write SetDesiredProcess;
    property DoRender: Boolean read FDoRender write SetDoRender;
    property DoProcess: Boolean read FDoProcess write SetDoProcess;
    property DoRealTime: Boolean read FDoRealTime write SetDoRealTime;
    property Enabled: Boolean read FEnabled write SetEnabled;
    property OnProcess: TNotifyEvent read FOnProcess write SetOnProcess;
    property OnRender: TNotifyEvent read FOnRender write SetOnRender;
    property OnRealTime: TRealTimeEvent read FOnRealTime write SetOnRealTime;
  end;

procedure Register;

implementation

{ TPreciseTimer }

procedure TPreciseTimer.ApplicationIdle(Sender: TObject;
  var Done: Boolean);
var
  Fps: Real;
begin
  if (csDesigning in ComponentState) then Exit;
  if not FEnabled then Exit;
  if (not FDoRealTime) and (not FDoRender) and (not FDoProcess) then Exit;

  Done := False;

  if FDoRender then begin
    QueryPerformanceCounter(FTimeCurrent);
    if FRenderLast = 0 then FRenderLast := FTimeCurrent;
    Fps := 1 / ((FTimeCurrent - FRenderLast - FTimeOverhead) / FTimeFreq);
    if ((FDesiredRender = 0) or (Fps <= FDesiredRender)) and Assigned(FOnRender) then begin
      FRenderLast := FTimeCurrent;
      FRealRender := Fps;
      FOnRender(Self);
    end;
  end;

  if FDoProcess then begin
    QueryPerformanceCounter(FTimeCurrent);
    if FProcessLast = 0 then FProcessLast := FTimeCurrent;
    Fps := 1 / ((FTimeCurrent - FProcessLast - FTimeOverhead) / FTimeFreq);
    if ((FDesiredProcess = 0) or (Fps <= FDesiredProcess)) and Assigned(FOnProcess) then begin
      FProcessLast := FTimeCurrent;
      FRealProcess := Fps;
      FOnProcess(Self);
    end;
  end;

  if FDoRealTime then begin
    QueryPerformanceCounter(FTimeCurrent);
    Fps := 1 / ((FTimeCurrent - FProcessLast - FTimeOverhead) / FTimeFreq);
    if Assigned(FOnRealTime) then
      FOnRealTime(Self,Fps);
  end;
end;

constructor TPreciseTimer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FEnabled := True;
  FDesiredRender := 0;
  FDesiredProcess := 60;

  if not (csDesigning in ComponentState) then begin
    QueryPerformanceFrequency(FTimeFreq);
    QueryPerformanceCounter(FTimeCurrent);
    QueryPerformanceCounter(FProcessLast);
    FTimeOverhead := FProcessLast - FTimeCurrent;
    FProcessLast := 0;
    Application.OnIdle := ApplicationIdle;
  end;
end;

destructor TPreciseTimer.Destroy;
begin
  if not (csDesigning in ComponentState) then
    Application.OnIdle := nil;
  inherited;
end;

procedure TPreciseTimer.SetDesiredProcess(const Value: Double);
begin
  FDesiredProcess := Value;
end;

procedure TPreciseTimer.SetDesiredRender(const Value: Double);
begin
  FDesiredRender := Value;
end;

procedure TPreciseTimer.SetDoProcess(const Value: Boolean);
begin
  FDoProcess := Value;
end;

procedure TPreciseTimer.SetDoRealTime(const Value: Boolean);
begin
  FDoRealTime := Value;
end;

procedure TPreciseTimer.SetDoRender(const Value: Boolean);
begin
  FDoRender := Value;
end;

procedure TPreciseTimer.SetEnabled(const Value: Boolean);
begin
  FEnabled := Value;
end;

procedure TPreciseTimer.SetOnProcess(const Value: TNotifyEvent);
begin
  FOnProcess := Value;
end;

procedure TPreciseTimer.SetOnRealTime(const Value: TRealTimeEvent);
begin
  FOnRealTime := Value;
end;

procedure TPreciseTimer.SetOnRender(const Value: TNotifyEvent);
begin
  FOnRender := Value;
end;

procedure Register;
begin
  RegisterComponents('Wetgos', [TPreciseTimer]);
end;

end.
 